mod cli;
mod quic;
mod req;

use std::{convert::Into, net::SocketAddr};

use clap::Parser as _;
use cli::Cli;
use lanspread_compat::eti;
use lanspread_db::db::{Game, GameDB};
use lanspread_mdns::{
    DaemonEvent, LANSPREAD_INSTANCE_NAME, LANSPREAD_SERVICE_TYPE, MdnsAdvertiser,
};
use quic::run_server;
use tracing_subscriber::EnvFilter;

fn spawn_mdns_task(server_addr: SocketAddr) -> eyre::Result<()> {
    let mdns = MdnsAdvertiser::new(LANSPREAD_SERVICE_TYPE, LANSPREAD_INSTANCE_NAME, server_addr)?;

    tokio::spawn(async move {
        while let Ok(event) = mdns.monitor.recv() {
            tracing::info!("mDNS: {:?}", &event);
            if let DaemonEvent::Error(e) = event {
                tracing::error!("mDNS: {e}");
                tokio::time::sleep(std::time::Duration::from_secs(1)).await;
                continue;
            }
        }
    });

    Ok(())
}

async fn prepare_game_db(cli: &Cli) -> eyre::Result<GameDB> {
    // build games from ETI database
    let mut games: Vec<Game> = eti::get_games(&cli.db)
        .await?
        .into_iter()
        .map(Into::into)
        .collect();

    // filter out games that the server does not have in game_dir
    games.retain(|game| cli.game_dir.join(&game.id).is_dir());

    let mut game_db = GameDB::from(games);

    game_db.add_thumbnails(&cli.thumbs_dir);

    tracing::info!("Prepared game database with {} games", game_db.games.len());

    Ok(game_db)
}

#[tokio::main]
async fn main() -> eyre::Result<()> {
    tracing_subscriber::fmt()
        .with_env_filter(EnvFilter::from_default_env())
        .init();

    let cli = Cli::parse();

    assert!(
        cli.game_dir.exists(),
        "Games folder does not exist: {}",
        cli.game_dir.to_str().expect("Invalid path")
    );

    let server_addr = SocketAddr::from((cli.ip, cli.port));

    // spawn mDNS listener task
    spawn_mdns_task(server_addr)?;

    let game_db = prepare_game_db(&cli).await?;

    tracing::info!("Server listening on {server_addr}");
    run_server(server_addr, game_db, cli.game_dir).await
}
